/*
Copyright (c) 2017, Cryst.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, 
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution.
    * Neither the name of the Cryst nor the names of its contributors may be 
      used to endorse or promote products derived from this software without 
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
OF THE POSSIBILITY OF SUCH DAMAGE.

Copyright (c)  2017, Cryst.
All rights reserved.

  ソースコード形式かバイナリ形式か、変更するかしないかを問わず、以下の条件を
  満たす場合に限り、再頒布および使用を許可します。

    * ソースコードを再頒布する場合、上記の著作権表示、本条件一覧、および 
      下記免責条項を含めること。
    * バイナリ形式で再頒布する場合、頒布物に付属のドキュメント等の資料に、
      上記の著作権表示、本条件一覧、および下記免責条項を含めること。
    * 書面による特別の許可なしに、本ソフトウェアから派生した製品の宣伝または
      販売促進に、Cryst の名前またはコントリビューターの名前を使用してはならない。

  本ソフトウェアは、著作権者およびコントリビューターによって「現状のまま」提供さ
  れており、明示黙示を問わず、商業的な使用可能性、および特定の目的に対する適合性
  に関する暗黙の保証も含め、またそれに限定されない、いかなる保証もありません。
  著作権者もコントリビューターも、事由のいかんを問わず、損害発生の原因いかんを問
  わず、かつ責任の根拠が契約であるか厳格責任であるか（過失その他の）不法行為であ
  るかを問わず、仮にそのような損害が発生する可能性を知らされていたとしても、
  本ソフトウェアの使用によって発生した（代替品または代用サービスの調達、使用の
  喪失、データの喪失、利益の喪失、業務の中断も含め、またそれに限定されない）
  直接損害、間接損害、偶発的な損害、特別損害、懲罰的損害、または結果損害について、
  一切責任を負わないものとします。
*/
////////////////////////////////////////////////////////////////////////////////
// コンパイル方法の例：
// gcc -O3 -Wall -W -o idft_2D idft_2D.c
////////////////////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define IMAGEPROCESSING              idft_2D
#define FINALFNAMEIS(foname)         #foname
#define EXTRACTING(foname)           FINALFNAMEIS(foname.bmp)
#define FNAMEIS(foname)              EXTRACTING(foname)
#define IFNAMEDEFAULT                "dft_complex.dat"

#define saveArray2bmp32(name,a,w,h)     saveDataAsBMP32((name),(&(a)[0][0]),(w),(h))
#define saveArray2COMPLEX(name,m,a,w,h) saveDataAsCOMPLEX((name),(m),(&(a)[0][0]),(w),(h))

#ifndef M_PI
#define M_PI            3.14159265358979323846
#endif

extern int  saveDataAsBMP32( char *, void *, int, int );
extern void **alloc2D( void *, int, int, int );
extern void **alloc2Darray( int, int, int );

#pragma pack(push,1)
typedef struct PIXEL   { unsigned char B, G, R, A; } PIXEL ;
#pragma pack(pop)

typedef struct COMPLEX { double r, i; } COMPLEX;
#define AddCOMPLEX(a,b)   (COMPLEX){ (a).r + (b).r, (a).i + (b).i }
#define SubCOMPLEX(a,b)   (COMPLEX){ (a).r - (b).r, (a).i - (b).i }
#define MulCOMPLEX(a,b)   (COMPLEX){ (a).r * (b).r - (a).i * (b).i, (a).r * (b).i + (a).i * (b).r }
#define DivCOMPLEX(a,b)   (COMPLEX){ ((a).r*(b).r + (a).i*(b).i)/((a).r*(a).r + (b).i*(b).i),\
                                     ((a).i*(b).r - (a).r*(b).i)/((a).r*(a).r + (b).i*(b).i) }
#define ImeMulCOMPLEX(a,b)(COMPLEX){ (a).r * (b), (a).i * (b) }
#define ImeDivCOMPLEX(a,b)(COMPLEX){ (a).r / (b), (a).i / (b) }
#define RevCOMPLEX(a)     (COMPLEX){ (a).r ,-((a).i) }
#define RotorCOMPLEX(a)   (COMPLEX){ cos(a), sin(a) }
#define ImeDivCX          ImeDivCOMPLEX

#pragma pack(push,2)
typedef struct HEADERCX { // COMPLEX => bfType==0x585043 (0x43,0x50,0x58)
    unsigned int   cxType, zero;    // identification 0x43='C',0x50='P',0x58='X'
    unsigned int   cxSize;          // ?? cxOffByts+cxSizeData+sizeof(message)
    unsigned int   cxDataByte;      // 16 sizeof(COMPLEX)
    unsigned int   cxOffByts;       // ?? 32+α sizeof(HEADER) +sizeof(message)
    unsigned int   cxSizeData;      // ?? lx*ly*sizeof(COMPLEX)
    signed   int   cxWidth;         // ?? lx
    signed   int   cxHeight;        // ?? ly
    ///////////////////////////////////////////////////////////////////////////
    char           message[];
} HEADERCX ;
#pragma pack(pop)
#define COMPLEXID     0x585043

enum DFT_DirectionOfConversion {
     DFT,
     IDFT
}    DFT_DirectionOfConversion_END;

void dft_TwoDimensional( COMPLEX **d, int lx, int ly, int flag ) {
int     x, y, u, v, elements;
double  unit;
COMPLEX t[(lx>ly?lx:ly)];

    elements = lx;  unit = (flag==IDFT) ? 2*M_PI/elements : -2*M_PI/elements;
    for( y=0; y<ly; y++ ) { // Σdft_oneDimensional( for Y axis ...);
        for( u=0; u<elements; u++ ) t[u]=(COMPLEX){0,0}; // From 0 To LX -1
        for( u=0; u<elements; u++ ) {
            for( v=0; v<elements; v++ ) t[u] = AddCOMPLEX( t[u], MulCOMPLEX(d[y][v],RotorCOMPLEX(unit*u*v)) );
        }
        if(flag==IDFT) for( u=0; u<elements; u++ ) d[y][u] = ImeDivCOMPLEX(t[u],elements);
        else           for( u=0; u<elements; u++ ) d[y][u] = t[u];
    } //////////////////////////////////////////////////////////////////////////
    elements = ly;  unit = (flag==IDFT) ? 2*M_PI/elements : -2*M_PI/elements;
    for( x=0; x<lx; x++ ) { // Σdft_oneDimensional( for X axis ...);
        for( u=0; u<elements; u++ ) t[u]=(COMPLEX){0,0}; // From 0 To LY -1
        for( u=0; u<elements; u++ ) {
            for( v=0; v<elements; v++ ) t[u] = AddCOMPLEX( t[u], MulCOMPLEX(d[v][x],RotorCOMPLEX(unit*u*v)) );
        }
        if(flag==IDFT) for( u=0; u<elements; u++ ) d[u][x] = ImeDivCOMPLEX(t[u],elements);
        else           for( u=0; u<elements; u++ ) d[u][x] = t[u];
    } //////////////////////////////////////////////////////////////////////////
}// See also, Euler's formula & Definition of trigonometric functions (Symmetry)

void complexReal2grayEx( COMPLEX **u, PIXEL **a, int lx, int ly ) {
double min, max, div;
int    hist[1024], lut[1024], **w=(int**)a, m, x, y;

    for( min=max=u[0][0].r,y=0; y<ly; y++ ) for( x=0; x<lx; x++ ) {
        if(      min>u[y][x].r ) min = u[y][x].r;
        else if( max<u[y][x].r ) max = u[y][x].r;
    }   div = max>min ? max-min : 1 ;
    if( 0<=min && max<256 ) {
        for( y=0; y<ly; y++ ) for( x=0; x<lx; x++ ) a[y][x]=(PIXEL){m=u[y][x].r,m,m,0};
    } else{
        for( x=0; x<1024; x++ ) hist[x]=lut[x]=0;
        for( y=0; y<ly; y++ ) for( x=0; x<lx; x++ ){
            hist[w[y][x]=1023.0*(u[y][x].r-min)/div]++;
        }
        int l, h, n, m=((lx*ly)/200)>0 ? ((lx*ly)/100) : 1 ;
        for(      n=l=0; (n+hist[l])<m && l<1024; l++ ) n+=hist[l];
        for( n=0,h=1023; (n+hist[h])<m && h>0;    h-- ) n+=hist[h];
    
        for( n=0; n<1024; n++ ) lut[n]=0xFFFFFF;
        for( n=0; n<=l;   n++ ) lut[n]=0;
        div = h>l ? h-l : 1 ;
        for( n=1; (l+n)<h; n++ ) { m=255*n/div; lut[l+n]= m<<16| m<<8 | m; }
        for( y=0; y<ly; y++ ) for( x=0; x<lx; x++ ) w[y][x]=lut[w[y][x]];
    }
}


int IMAGEPROCESSING( COMPLEX **s, PIXEL **d, int lx, int ly ){

    // 逆フーリエ変換（周波数表現 => 画像） 四隅が低周波、中心が高周波成分
    dft_TwoDimensional( s, lx, ly, IDFT );
    complexReal2grayEx( s, d, lx, ly );
    return 0 ;
}
////////////////////////////////////////////////////////////////////////////////
/// 以下、メインプログラム関連ルーチンとBMP画像を処理する共通ルーチン群
///
#include <sys/stat.h>

#define BMPID      0x4D42
#define BMPIFHSIZE 40

#pragma pack(push,2)
typedef struct HEADER {       // BMP => bfType==0x4D42 (0x42,0x4D) little endian
    unsigned short bfType;          // identification  0x42='B', 0x4D='M'
    unsigned int   bfSize;          // ?? 
    unsigned short bfReserved1;     //  0 
    unsigned short bfReserved2;     //  0 
    unsigned int   bfOffBits;       // 54  sizeof(HEADER)
    ////////////////////////////////////////////////////////////////////////////
    unsigned int   biSize;          // 40 sizeof(biSize ～ biClrImportant)
    signed int     biWidth;         // lx
    signed int     biHeight;        // ly
    unsigned short biPlanes;        //  1
    unsigned short biBitCount;      // 32
    unsigned int   biCompression;   //  0
    unsigned int   biSizeImage;     // ??
    signed int     biXPelsPerMeter; //  0
    signed int     biYPelsPerMeter; //  0
    unsigned int   biClrUsed;       //  0
    unsigned int   biClrImportant;  //  0
    ////////////////////////////////////////////////////////////////////////////
    unsigned char  lut[];
} HEADER ;
#pragma pack(pop)

void **alloc2D( void *data, int s, int lx, int ly ){
// データ（lx*ly個）の１次元配列を、array[y][x] 形式で参照できるようにする
void **d, *base=data ;
long i;

    if( (data==NULL) || (s<1) || (lx<1) || (ly<1) ) return NULL ;
    if( (d=(void **)malloc(ly*sizeof(void *))) == NULL ) return NULL ;
    for( i=0; i<ly; i++ ) d[i] = (void *)((unsigned char *)base + lx*s*i) ;
    return d ;
}

#define calloc2Darray alloc2Darray
void **alloc2Darray( int s, int lx, int ly ){
// array[y][x] 形式で参照できる、横：lx、縦：ly の配列を動的に確保する
void **d, *base ;
long i ;

    if( (s<1) || (lx<1) || (ly<1) ) return NULL ;
    if( (d=(void **)calloc(1,ly*sizeof(void *)+(lx*ly*s))) == NULL ) return NULL ;
    base = (unsigned char *)d + ly*sizeof(void *) ;
    for( i=0; i<ly; i++ ) d[i] = (void *)((unsigned char *)base + lx*s*i) ;
    return d ;
}

int saveDataAsBMP32( char *fname, void *b, int lx, int ly ){
// 処理結果を 32ビット形式の BMPファイルとして書き出す。
HEADER h = {BMPID, lx*abs(ly)*sizeof(PIXEL)+sizeof(HEADER), 0, 0, sizeof(HEADER),
            BMPIFHSIZE, lx, ly, 1, 32, 0, lx*abs(ly)*sizeof(PIXEL), 0, 0, 0, 0} ;
FILE *fp ;

    if( (fp=fopen(fname, "wb")) == NULL ) {
        fprintf( stderr, "File '%s' cannot create.\n", fname );
        return -1;
    } else if ( fwrite( &h, sizeof(HEADER), 1, fp ) != 1 ) {
        fclose(fp) ; return -1;
    } else if ( fwrite( b, h.biSizeImage, 1, fp ) != 1 ) {
        fclose(fp) ; return -1;
    } else fclose(fp) ;
    return 0;
}

void *dataFile2memory( char *fname ){
// ディスク上のファイルを、メモリーに一括で読み込む
struct stat fileInfo ;
FILE   *fp ;
char   *raw=NULL ;

    if ( (fp=fopen(fname,"rb")) == NULL ) {
        fprintf( stderr, "'%s' open error.\n", fname );               return raw;
    } else if ( stat(fname, &fileInfo) != 0 ) {
        fprintf( stderr, "'%s' stat error.\n", fname );  fclose(fp);  return raw;
    } else if( (raw=(char *)malloc(fileInfo.st_size+32)) == NULL ) {
        fprintf( stderr, "Couldn't allocate memory.\n" ) ; fclose(fp);return raw;
    } else if ( fread(raw, fileInfo.st_size, 1, fp) != 1 ) {
        fprintf( stderr, "'%s' read error.\n", fname );
        free(raw); fclose(fp);  return NULL;
    } else fclose(fp) ;
    return raw;
}

int check_complex_array( void *raw, unsigned int size ) {
HEADERCX *h=(HEADERCX*)raw;

    if( size<(int)sizeof(HEADERCX) )     return 1;
    if( h->cxType!=COMPLEXID ) return 1;
    if( h->cxWidth<=1 || h->cxHeight<=1 )           return 1;
    return (sizeof(HEADERCX)+h->cxWidth*h->cxHeight*sizeof(COMPLEX))>size ? 1 : 0;
}

void argumentAnalysis(int argc, char *argv[], char **iname, char **oname){
int  i, st=1;   // プログラム起動時の引数解析処理
    for( i=1; i<argc; i++ ) {
        if( argv[i][0]=='-' ) {
            switch(argv[i][1]) {
                case 'h': st=0;
                default : goto GOOD_BY;
            }
        }
        else if( *iname==NULL ) *iname = argv[i];
        else if( *oname==NULL ) *oname = argv[i];
        else goto GOOD_BY;
    }
    return;
GOOD_BY:;
    char *p=argv[0], *op="[-h] ";
    for( i=strlen(p); i>0 && p[i]!='\\' && p[i]!='/'; i-- );; if(i) i++;
    fprintf( stderr, "\nusage: %s %s[complex_in.dat [output_bmp]]\n\n", &p[i], op );
    fprintf( stderr, "\t-h: This option shows only this message.\n" );
    exit(st);
}

int main(int argc, char *argv[]){
int     st=1;
char    *raw=NULL, *ifname=NULL, *ofname=NULL;
PIXEL   **destination=NULL ;
COMPLEX **src=NULL;

    argumentAnalysis( argc, argv, &ifname, &ofname );    // 引数解析
    if( ifname==NULL ) ifname = IFNAMEDEFAULT;           // 入力ファイル名の設定
    if( ofname==NULL ) ofname = FNAMEIS(IMAGEPROCESSING);// 出力ファイル名の設定
    
    // データファイルの確認
    struct stat  fst ;
    if ( (st=stat( ifname, &fst )) ) {
        fprintf(stderr, "'%s' stat error.\n", ifname );  goto GOOD_BY;
    } else if( fst.st_size<(int)sizeof(HEADERCX) ) {
        fprintf(stderr, "'%s' size error.\n", ifname );  goto GOOD_BY;
    }
    // データの読み込み（ディスクからメモリーへ）
    if( (raw=(char*)dataFile2memory(ifname))==NULL ) {
        fprintf(stderr, "'%s' read error.\n", ifname );  goto GOOD_BY;
    }
    // データの確認
    if( (st=check_complex_array( raw, fst.st_size )) ) {
        fprintf(stderr, "'%s' unknown file\n", ifname ); goto GOOD_BY;
    }

    HEADERCX *h = (HEADERCX*)raw;
    COMPLEX  *c = (COMPLEX*)&((char *)raw)[h->cxOffByts];
    int      lx = h->cxWidth,  ly = h->cxHeight;
    // 作業用環境の構築
    if (  (destination=(PIXEL**)alloc2Darray(sizeof(PIXEL),lx,ly)) == NULL ) {
        fprintf(stderr, "Couldn't allocate memory. destination\n");goto GOOD_BY;
    } else if( (src=(COMPLEX**)alloc2D(c,sizeof(COMPLEX),lx,ly)) == NULL ) {
        fprintf(stderr, "Couldn't allocate memory. source\n");     goto GOOD_BY;
    }

    // 実際の処理ルーチンの呼び出し
    if( (st=IMAGEPROCESSING( src, destination, lx, ly ))==0 ) {
        saveArray2bmp32( ofname, destination, lx, ly );  // BMPファイルで保存
    } else if( st>0 ) fprintf( stderr, "Image processing is not completed.\n" );
    
GOOD_BY:; // 後始末とプログラムの終了
    free(src); free(destination); free(raw);
    return st;
}
